//
//  NSObject+HHAssociatedObjects.m
//
// Copyright (c) 2010 Houdah Software s.à r.l. (http://www.houdah.com)
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use,
// copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following
// conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//

#import "NSObject+HHAssociatedObjects.h"

#import <libkern/OSAtomic.h>
#import <objc/runtime.h>


@implementation NSObject (HHAssociatedObjects)


static inline BOOL associateValueSupportAvailable()
{
	static int associateValueSupportAvailableInt = -1;
	
	if (associateValueSupportAvailableInt == -1) {
		associateValueSupportAvailableInt = ((objc_setAssociatedObject != NULL) && (objc_getAssociatedObject != NULL));
	}
	
	return associateValueSupportAvailableInt;
}

+ (NSMapTable*)associateValuesByObject
{
	static NSMapTable *associateValuesByObject = nil;
	
	if (associateValuesByObject == nil) {
		NSMapTable *mapTable = [[NSMapTable alloc] initWithKeyOptions:NSMapTableObjectPointerPersonality
														 valueOptions:NSMapTableStrongMemory
															 capacity:100];

		if (!OSAtomicCompareAndSwapPtrBarrier(nil, mapTable, (void *)&associateValuesByObject)) {
			[mapTable release];
		}
	}
	
	return associateValuesByObject;
}

+ (void)swizzleRelease
{
	static BOOL swizzleReleaseDone = NO;
	
	if (! swizzleReleaseDone) {
		@synchronized ([self class]) {
			if (! swizzleReleaseDone) {
				SEL origSel = @selector(release);
				SEL altSel = @selector(releaseAssociatedValues);
				Method origMethod = class_getInstanceMethod(self, origSel);
				Method altMethod = class_getInstanceMethod(self, altSel);

				class_addMethod(self,
								origSel,
								class_getMethodImplementation(self, origSel),
								method_getTypeEncoding(origMethod));
				class_addMethod(self,
								altSel,
								class_getMethodImplementation(self, altSel),
								method_getTypeEncoding(altMethod));
				
				method_exchangeImplementations(class_getInstanceMethod(self, origSel), class_getInstanceMethod(self, altSel));
				
				swizzleReleaseDone = YES;
			}
		}
	}
}


- (NSMutableDictionary*)associateValues
{
	NSMapTable *associateValuesByObject = [[self class] associateValuesByObject];
	NSMutableDictionary *associateValues = [associateValuesByObject objectForKey:self];
	
	if (associateValues == nil) {		
		@synchronized (self) {
			associateValues = [associateValuesByObject objectForKey:self];

			if (associateValues == nil) {
				NSMutableDictionary *dictionary = [NSMutableDictionary dictionaryWithCapacity:5];

				[associateValuesByObject setObject:dictionary forKey:self];

				[[self class] swizzleRelease];
				
				associateValues = dictionary;
			}
		}
	}
	
	return associateValues;
}
 
- (void)storeAssociateValue:(id)value forKey:(void *)key
{
	NSMutableDictionary *associateValues = [self associateValues];
	
	if (value != nil) {
		[associateValues setObject:value forKey:key];
	}
	else {
		[associateValues removeObjectForKey:key];
	}
}

- (id)retrieveAssociateValueForKey:(void *)key
{
	NSMutableDictionary *associateValues = [self associateValues];

	return [associateValues objectForKey:key];
}

- (void)associateValue:(id)value withKey:(void *)key
{
	if (associateValueSupportAvailable()) {
		objc_setAssociatedObject(self, key, value, OBJC_ASSOCIATION_RETAIN);
	}
	else {
		[self storeAssociateValue:(id)value forKey:(void *)key];
	}
}

- (id)associatedValueForKey:(void *)key
{	
	if (associateValueSupportAvailable()) {		
		return objc_getAssociatedObject(self, key);
	}
	else {
		return [self retrieveAssociateValueForKey:(void *)key];
	}	
}

- (void)releaseAssociatedValues
{
	NSUInteger retainCount = [self retainCount];

	[self releaseAssociatedValues];
	
	if (retainCount == 2) {
		NSMapTable *associateValuesByObject = [[self class] associateValuesByObject];
				
		[associateValuesByObject removeObjectForKey:self];
	}
}

@end